import deprecation
from .domain import C4k, DnsRecord
from .devops_build import DevopsBuild
from .credential import gopass_field_from_path, gopass_password_from_path
from .infrastructure import ProjectRepository, FileApi, ExecutionApi


@deprecation.deprecated(deprecated_in="3.2")
# create objects direct instead
def add_c4k_mixin_config(
    config,
    c4k_config_dict,
    c4k_auth_dict,
    executabel_name=None,
    grafana_cloud_user=None,
    grafana_cloud_password=None,
    grafana_cloud_url="https://prometheus-prod-01-eu-west-0.grafana.net/api/prom/push",
):
    if not grafana_cloud_user:
        grafana_cloud_user = gopass_field_from_path(
            "server/meissa/grafana-cloud", "grafana-cloud-user"
        )
    if not grafana_cloud_password:
        grafana_cloud_password = gopass_password_from_path(
            "server/meissa/grafana-cloud"
        )
    c4k_auth_dict.update(
        {
            "mon-auth": {
                "grafana-cloud-user": grafana_cloud_user,
                "grafana-cloud-password": grafana_cloud_password,
            }
        }
    )
    c4k_config_dict.update({"mon-cfg": {"grafana-cloud-url": grafana_cloud_url}})
    config.update(
        {
            "C4kMixin": {
                "executabel_name": executabel_name,
                "config": c4k_config_dict,
                "auth": c4k_auth_dict,
            }
        }
    )
    return config


#TODO: refactor this to C4kBuild
class C4kMixin(DevopsBuild):
    def __init__(self, project, config):
        super().__init__(project, config)
        self.execution_api = ExecutionApi()
        c4k_build = C4k(config)
        self.repo.set_c4k(self.project, c4k_build)

    def update_runtime_config(self, dns_record: DnsRecord):
        c4k_build = self.repo.get_c4k(self.project)
        c4k_build.update_runtime_config(dns_record)
        self.repo.set_c4k(self.project, c4k_build)

    def write_c4k_config(self):
        build = self.repo.get_devops(self.project)
        c4k_build = self.repo.get_c4k(self.project)
        path = build.build_path() + "/out_c4k_config.yaml"
        self.file_api.write_yaml_to_file(path, c4k_build.config())

    def write_c4k_auth(self):
        build = self.repo.get_devops(self.project)
        c4k_build = self.repo.get_c4k(self.project)
        path = build.build_path() + "/out_c4k_auth.yaml"
        self.file_api.write_yaml_to_file(path, c4k_build.c4k_mixin_auth)

    def c4k_apply(self, dry_run=False):
        build = self.repo.get_devops(self.project)
        c4k_build = self.repo.get_c4k(self.project)
        return self.execution_api.execute(c4k_build.command(build), dry_run)



from .infrastructure import ReleaseRepository
from .infrastructure_api import GitApi
from .domain import Version, Release

class PrepareReleaseService():

    def __init__(self, release_repo: ReleaseRepository):
        self.release_repo = release_repo
        self.release = release_repo.get_release()
        self.git_api = GitApi()

    def __write_and_commit_version(self, version: Version, commit_message: str):
        self.release.validate(self.release_repo.main_branch)

        self.release_repo.version_repository.write_file(version.get_version_string())
        self.git_api.add_file(self.release_repo.version_repository.file)
        self.git_api.commit(commit_message)

    def write_and_commit_release(self):
        self.__write_and_commit_version(self.release.release_version(), commit_message=f'Release v{self.release.release_version().get_version_string()}')

    def write_and_commit_bump(self):
        self.__write_and_commit_version(self.release.bump_version(), commit_message='Version bump')

class TagAndPushReleaseService():

    def __init__(self, git_api: GitApi):
        self.git_api = git_api

    def tag_release(self, release: Release):
        annotation = 'v' + release.version.get_version_string()
        message = 'Release ' + annotation
        self.git_api.tag_annotated(annotation, message, 1)

    def push_release(self):
        self.git_api.push()



"""
ddadevops provide tools to support builds combining gopass,
terraform, dda-pallet, aws & hetzner-cloud.

"""

from .python_util import execute
from .provs_k3s_mixin import ProvsK3sMixin, add_provs_k3s_mixin_config
from .aws_rds_pg_mixin import AwsRdsPgMixin, add_aws_rds_pg_mixin_config
from .aws_mfa_mixin import AwsMfaMixin, add_aws_mfa_mixin_config
from .aws_backend_properties_mixin import AwsBackendPropertiesMixin, add_aws_backend_properties_mixin_config
from .c4k_mixin import C4kMixin, add_c4k_mixin_config
from .exoscale_mixin import ExoscaleMixin, add_exoscale_mixin_config
from .digitalocean_backend_properties_mixin import DigitaloceanBackendPropertiesMixin, add_digitalocean_backend_properties_mixin_config
from .digitalocean_terraform_build import DigitaloceanTerraformBuild, create_digitalocean_terraform_build_config
from .hetzner_mixin import HetznerMixin, add_hetzner_mixin_config
from .devops_docker_build import DevopsDockerBuild, create_devops_docker_build_config
from .devops_terraform_build import DevopsTerraformBuild, create_devops_terraform_build_config
from .devops_build import DevopsBuild, create_devops_build_config, get_devops_build, get_tag_from_latest_commit
from .credential import gopass_password_from_path, gopass_field_from_path

from .domain import Validateable, DnsRecord, Devops, Docker, C4k
from .application import DockerBuildService
from .infrastructure import ProjectRepository, ResourceApi, FileApi, DockerApi, ExecutionApi

__version__ = "${version}"



import sys
from os import chmod
from json import load, dumps
from subprocess import run
from packaging import version
from pkg_resources import resource_string
from dda_python_terraform import Terraform, IsFlagged
from .python_util import filter_none
from .devops_build import DevopsBuild, create_devops_build_config




def create_devops_terraform_build_config(stage,
                                         project_root_path,
                                         module,
                                         additional_vars,
                                         build_dir_name='target',
                                         output_json_name=None,
                                         use_workspace=True,
                                         use_package_common_files=True,
                                         build_commons_path=None,
                                         terraform_build_commons_dir_name='terraform',
                                         debug_print_terraform_command=False,
                                         additional_tfvar_files=None,
                                         terraform_semantic_version="1.0.8"):
    if not output_json_name:
        output_json_name = 'out_' + module + '.json'
    if not additional_tfvar_files:
        additional_tfvar_files = []
    ret = create_devops_build_config(
        stage, project_root_path, module, build_dir_name)
    ret.update({'additional_vars': additional_vars,
                'output_json_name': output_json_name,
                'use_workspace': use_workspace,
                'use_package_common_files': use_package_common_files,
                'build_commons_path': build_commons_path,
                'terraform_build_commons_dir_name': terraform_build_commons_dir_name,
                'debug_print_terraform_command': debug_print_terraform_command,
                'additional_tfvar_files': additional_tfvar_files,
                'terraform_semantic_version': terraform_semantic_version})
    return ret

class DevopsTerraformBuild(DevopsBuild):

    def __init__(self, project, config):
        super().__init__(project, config)
        project.build_depends_on('dda-python-terraform')
        self.additional_vars = config['additional_vars']
        self.output_json_name = config['output_json_name']
        self.use_workspace = config['use_workspace']
        self.use_package_common_files = config['use_package_common_files']
        self.build_commons_path = config['build_commons_path']
        self.terraform_build_commons_dir_name = config['terraform_build_commons_dir_name']
        self.debug_print_terraform_command = config['debug_print_terraform_command']
        self.additional_tfvar_files = config['additional_tfvar_files']
        self.terraform_semantic_version = config['terraform_semantic_version']

    def terraform_build_commons_path(self):
        mylist = [self.build_commons_path,
                  self.terraform_build_commons_dir_name]
        return '/'.join(filter_none(mylist)) + '/'

    def project_vars(self):
        ret = {'stage': self.stage}
        if self.module:
            ret['module'] = self.module
        if self.additional_vars:
            ret.update(self.additional_vars)
        return ret

    def copy_build_resource_file_from_package(self, name):
        my_data = resource_string(
            __name__, "src/main/resources/terraform/" + name)
        with open(self.build_path() + '/' + name, "w", encoding="utf-8") as output_file:
            output_file.write(my_data.decode(sys.stdout.encoding))

    def copy_build_resources_from_package(self):
        self.copy_build_resource_file_from_package('versions.tf')
        self.copy_build_resource_file_from_package('terraform_build_vars.tf')

    def copy_build_resources_from_dir(self):
        run('cp -f ' + self.terraform_build_commons_path() +
            '* ' + self.build_path(), shell=False, check=False)

    def copy_local_state(self):
        run('cp terraform.tfstate '  + self.build_path(), shell=False, check=False)

    def rescue_local_state(self):
        run('cp ' + self.build_path() + '/terraform.tfstate .', shell=False, check=False)

    def initialize_build_dir(self):
        super().initialize_build_dir()
        if self.use_package_common_files:
            self.copy_build_resources_from_package()
        else:
            self.copy_build_resources_from_dir()
        self.copy_local_state()
        run('cp *.tf ' + self.build_path(), shell=True, check=False)
        run('cp *.properties ' + self.build_path(), shell=True, check=False)
        run('cp *.tfvars ' + self.build_path(), shell=True, check=False)
        run('cp -r scripts ' + self.build_path(), shell=True, check=False)

    def post_build(self):
        self.rescue_local_state()

    def init_client(self):
        terraform = Terraform(working_dir=self.build_path(), terraform_semantic_version=self.terraform_semantic_version)
        terraform.init()
        self.print_terraform_command(terraform)
        if self.use_workspace:
            try:
                terraform.workspace('select', self.stage)
                self.print_terraform_command(terraform)
            except:
                terraform.workspace('new', self.stage)
                self.print_terraform_command(terraform)
        return terraform

    def write_output(self, terraform):
        result = terraform.output(json=IsFlagged)
        self.print_terraform_command(terraform)
        with open(self.build_path() + self.output_json_name, "w", encoding="utf-8") as output_file:
            output_file.write(dumps(result))
        chmod(self.build_path() + self.output_json_name, 0o600)

    def read_output_json(self):
        with open(self.build_path() + self.output_json_name, 'r', encoding="utf-8") as file:
            return load(file)

    def plan(self):
        terraform = self.init_client()
        return_code, _, stderr = terraform.plan(detailed_exitcode=None, capture_output=False, raise_on_error=False,
                var=self.project_vars(),
                var_file=self.additional_tfvar_files)
        self.post_build()
        self.print_terraform_command(terraform)
        if return_code > 0:
            raise Exception(return_code, "terraform error:", stderr)

    def plan_fail_on_diff(self):
        terraform = self.init_client()
        return_code, _, stderr = terraform.plan(detailed_exitcode=IsFlagged, capture_output=False, raise_on_error=False,
                var=self.project_vars(),
                var_file=self.additional_tfvar_files)
        self.post_build()
        self.print_terraform_command(terraform)
        if return_code not in (0, 2):
            raise Exception(return_code, "terraform error:", stderr)
        if return_code == 2:
            raise Exception(return_code, "diff in config found:", stderr)

    def apply(self, auto_approve=False):
        terraform = self.init_client()
        if auto_approve:
            auto_approve_flag = IsFlagged
        else:
            auto_approve_flag = None
        if version.parse(self.terraform_semantic_version) >= version.parse("1.0.0"):
            return_code, _, stderr = terraform.apply(capture_output=False, raise_on_error=True,
                    auto_approve=auto_approve_flag,
                    var=self.project_vars(),
                    var_file=self.additional_tfvar_files)
        else:
            return_code, _, stderr = terraform.apply(capture_output=False, raise_on_error=True,
                    skip_plan=auto_approve,
                    var=self.project_vars(),
                    var_file=self.additional_tfvar_files)
        self.write_output(terraform)
        self.post_build()
        self.print_terraform_command(terraform)
        if return_code > 0:
            raise Exception(return_code, "terraform error:", stderr)

    def refresh(self):
        terraform = self.init_client()
        return_code, _, stderr = terraform.refresh(
                var=self.project_vars(),
                var_file=self.additional_tfvar_files)
        self.write_output(terraform)
        self.post_build()
        self.print_terraform_command(terraform)
        if return_code > 0:
            raise Exception(return_code, "terraform error:", stderr)

    def destroy(self, auto_approve=False):
        terraform = self.init_client()
        if auto_approve:
            auto_approve_flag = IsFlagged
        else:
            auto_approve_flag = None
        if version.parse(self.terraform_semantic_version) >= version.parse("1.0.0"):
            return_code, _, stderr = terraform.destroy(capture_output=False, raise_on_error=True,
                    auto_approve=auto_approve_flag,
                    var=self.project_vars(),
                    var_file=self.additional_tfvar_files)
        else:
            return_code, _, stderr = terraform.destroy(capture_output=False, raise_on_error=True,
                    force=auto_approve_flag,
                    var=self.project_vars(),
                    var_file=self.additional_tfvar_files)
        self.post_build()
        self.print_terraform_command(terraform)
        if return_code > 0:
            raise Exception(return_code, "terraform error:", stderr)

    def tf_import(self, tf_import_name, tf_import_resource,):
        terraform = self.init_client()
        return_code, _, stderr = terraform.import_cmd(tf_import_name, tf_import_resource,
                      capture_output=False, raise_on_error=True,
                      var=self.project_vars(),
                      var_file=self.additional_tfvar_files)
        self.post_build()
        self.print_terraform_command(terraform)
        if return_code > 0:
            raise Exception(return_code, "terraform error:", stderr)

    def print_terraform_command(self, terraform):
        if self.debug_print_terraform_command:
            output = 'cd ' + self.build_path() + ' && ' + terraform.latest_cmd()
            print(output)



from ddadevops import DevopsBuild
from .infrastructure import ReleaseRepository, ReleaseTypeRepository, VersionRepository
from .infrastructure_api import GitApi
from .services import PrepareReleaseService, TagAndPushReleaseService

def create_release_mixin_config(config_file, main_branch) -> dict:
    config = {}
    config.update(
        {'ReleaseMixin':
            {'main_branch': main_branch,
             'config_file': config_file}})
    return config

def add_versions(config, release_version, bump_version) -> dict:
    config['ReleaseMixin'].update(
        {'release_version': release_version,
         'bump_version': bump_version})
    return config

class ReleaseMixin(DevopsBuild):

    def __init__(self, project, config): # todo: create services in init, dont expose repos etc in api
        super().__init__(project, config)
        release_mixin_config = config['ReleaseMixin']                
        self.config_file = release_mixin_config['config_file']
        self.main_branch = release_mixin_config['main_branch']
        self.git_api = GitApi()
        self.release_type_repo = ReleaseTypeRepository(self.git_api) # maybe get from env?
        self.version_repo = VersionRepository(self.config_file)
        self.release_repo = ReleaseRepository(self.version_repo, self.release_type_repo, self.main_branch)

    def prepare_release(self):
        prepare_release_service = PrepareReleaseService(self.release_repo)
        prepare_release_service.write_and_commit_release()
        prepare_release_service.write_and_commit_bump()

    def tag_and_push_release(self):
        tag_and_push_release_service = TagAndPushReleaseService(self.git_api)
        tag_and_push_release_service.tag_release(self.release_repo.get_release())
        # tag_and_push_release_service.push_release()



import sys
import os
from pathlib import Path
from ddadevops import *

# getting the name of the directory
# where the this file is present.
current = os.path.dirname(os.path.realpath(__file__))
 
# adding the current directory to
# the sys.path.
sys.path.append(current)
 
# now we can import the module in the current
# directory.

from pybuilder.core import task, init
from ddadevops import *
from release_mixin import ReleaseMixin, create_release_mixin_config

CONFIG_FILE = Path('config.json')
MAIN_BRANCH = 'main'
STAGE = 'test'
PROJECT_ROOT_PATH = '.'
MODULE = 'test'
BUILD_DIR_NAME = "build_dir"

class MyBuild(ReleaseMixin):
    pass

@init
def initialize(project):
    project.build_depends_on('ddadevops>=3.1.2')
    config = create_release_mixin_config(CONFIG_FILE, MAIN_BRANCH)
    config.update({'stage': STAGE,
                   'module': MODULE,
                   'project_root_path': PROJECT_ROOT_PATH,
                   'build_dir_name': BUILD_DIR_NAME})
    build = MyBuild(project, config)
    build.initialize_build_dir()

@task
def release(project):
    build = get_devops_build(project)

    build.prepare_release()
    build.tag_and_push_release()



import deprecation
from typing import List
from .python_util import filter_none


class Validateable:
    def __validate_is_not_empty__(self, field_name: str) -> List[str]:
        value = self.__dict__[field_name]
        if value is None or value == "":
            return [f"Field '{field_name}' may not be empty."]
        else:
            return []

    def validate(self) -> List[str]:
        return []

    def is_valid(self) -> bool:
        return len(self.validate()) < 1


class DnsRecord(Validateable):
    def __init__(self, fqdn, ipv4=None, ipv6=None):
        self.fqdn = fqdn
        self.ipv4 = ipv4
        self.ipv6 = ipv6

    def validate(self) -> List[str]:
        result = []
        result += self.__validate_is_not_empty__("fqdn")
        if (not self.ipv4) and (not self.ipv6):
            result.append("ipv4 & ipv6 may not both be empty.")
        return result


class Devops(Validateable):
    def __init__(
        self, stage, project_root_path, module, name=None, build_dir_name="target"
    ):
        self.stage = stage
        self.name = name
        self.project_root_path = project_root_path
        self.module = module
        if not name:
            self.name = module
        self.build_dir_name = build_dir_name
        # Deprecated - no longer use generic stack ...
        self.stack = {}

    @deprecation.deprecated(deprecated_in="3.2")
    # use .name instead
    def name(self):
        return self.name

    def build_path(self):
        path = [self.project_root_path, self.build_dir_name, self.name, self.module]
        return "/".join(filter_none(path))

    def __put__(self, key, value):
        self.stack[key] = value

    def __get__(self, key):
        return self.stack[key]

    def __get_keys__(self, keys):
        result = {}
        for key in keys:
            result[key] = self.__get__(key)
        return result


class Docker(Validateable):
    def __init__(
        self,
        dockerhub_user,
        dockerhub_password,
        devops: Devops,
        build_dir_name="target",
        use_package_common_files=True,
        build_commons_path=None,
        docker_build_commons_dir_name="docker",
        docker_publish_tag=None,
    ):
        self.dockerhub_user = dockerhub_user
        self.dockerhub_password = dockerhub_password
        self.use_package_common_files = use_package_common_files
        self.build_commons_path = build_commons_path
        self.docker_build_commons_dir_name = docker_build_commons_dir_name
        self.docker_publish_tag = docker_publish_tag
        self.devops = devops

    def docker_build_commons_path(self):
        list = [self.build_commons_path, self.docker_build_commons_dir_name]
        return "/".join(filter_none(list)) + "/"


class C4k(Validateable):
    def __init__(self, config: map):
        tmp_executabel_name = config["C4kMixin"]["executabel_name"]
        if not tmp_executabel_name:
            tmp_executabel_name = config["module"]
        self.executabel_name = tmp_executabel_name
        self.c4k_mixin_config = config["C4kMixin"]["config"]
        self.c4k_mixin_auth = config["C4kMixin"]["auth"]
        tmp = self.c4k_mixin_config["mon-cfg"]
        tmp.update({"cluster-name": config["module"], "cluster-stage": config["stage"]})
        self.c4k_mixin_config.update({"mon-cfg": tmp})
        self.dns_record = None

    # TODO: these functions should be located at TerraformBuild later on.
    def update_runtime_config(self, dns_record: DnsRecord):
        self.dns_record = dns_record

    def validate(self) -> List[str]:
        result = []
        result += self.__validate_is_not_empty__("fqdn")
        if self.dns_record:
            result += self.dns_record.validate()
        return result

    def config(self):
        fqdn = self.dns_record.fqdn
        self.c4k_mixin_config.update({"fqdn": fqdn})
        return self.c4k_mixin_config

    def command(self, build: Devops):
        module = build.module
        build_path = build.build_path()
        config_path = f"{build_path}/out_c4k_config.yaml"
        auth_path = f"{build_path}/out_c4k_auth.yaml"
        output_path = f"{build_path}/out_{module}.yaml"
        return f"c4k-{self.executabel_name}-standalone.jar {config_path} {auth_path} > {output_path}"



from .domain import Release, Version, ReleaseType
from .infrastructure_api import FileHandler

class VersionRepository():

    def __init__(self, file):
        self.file = file
        self.file_handler = None
    
    def load_file(self):
        self.file_handler = FileHandler.from_file_path(self.file)
        return self.file_handler

    def write_file(self, version_string):
        if self.file_handler is None:
            raise Exception('Version was not created by load_file method.')        
        else:
            self.file_handler.write(version_string)
    
    def parse_file(self):
        version_list, is_snapshot = self.file_handler.parse()
        return version_list, is_snapshot
    
    def get_version(self) -> Version:

        self.file_handler = self.load_file()
        version_list, is_snapshot = self.parse_file()
        version = Version(self.file, version_list)
        version.is_snapshot = is_snapshot
        
        return version

class ReleaseTypeRepository():
    def __init__(self, git_api, environment_api=None):
        self.git_api = git_api

    def get_release_type(self):
        latest_commit = self.git_api.get_latest_commit()

        if ReleaseType.MAJOR.name in latest_commit.upper():
            return ReleaseType.MAJOR
        elif ReleaseType.MINOR.name in latest_commit.upper():
            return ReleaseType.MINOR
        elif ReleaseType.PATCH.name in latest_commit.upper():
            return ReleaseType.PATCH
        elif ReleaseType.SNAPSHOT.name in latest_commit.upper():
            return ReleaseType.SNAPSHOT
        else:
            return None

class ReleaseRepository():
    def __init__(self, version_repository: VersionRepository, release_type_repository: ReleaseTypeRepository, main_branch: str):
        self.version_repository = version_repository
        self.release_type_repository = release_type_repository
        self.main_branch = main_branch

    def get_release(self) -> Release:
        return Release(self.release_type_repository.get_release_type(), self.version_repository.get_version(), self.main_branch)



from enum import Enum
from pathlib import Path

class ReleaseType(Enum):
    MAJOR = 0
    MINOR = 1
    PATCH = 2
    SNAPSHOT = 3
    BUMP = None

class Version():

    def __init__(self, id: Path, version_list: list):
        self.id = id
        self.version_list = version_list
        self.version_string = None
        self.is_snapshot = None

    def increment(self, release_type: ReleaseType):
        self.is_snapshot = False
        match release_type:
            case ReleaseType.BUMP:
                self.is_snapshot = True
                self.version_list[ReleaseType.PATCH.value] += 1
            case ReleaseType.SNAPSHOT:
                self.is_snapshot = True
            case ReleaseType.PATCH:
                self.version_list[ReleaseType.PATCH.value] += 1
            case ReleaseType.MINOR:
                self.version_list[ReleaseType.PATCH.value] = 0
                self.version_list[ReleaseType.MINOR.value] += 1
            case ReleaseType.MAJOR:
                self.version_list[ReleaseType.PATCH.value] = 0
                self.version_list[ReleaseType.MINOR.value] = 0
                self.version_list[ReleaseType.MAJOR.value] += 1
            case None:
                raise Exception("Release Type was not set!")

    def get_version_string(self) -> str:
        self.version_string = ".".join([str(x) for x in self.version_list])
        if self.is_snapshot:
            self.version_string += "-SNAPSHOT"
        return self.version_string

    def create_release_version(self, release_type: ReleaseType):
        release_version = Version(self.id, self.version_list.copy())
        release_version.is_snapshot = self.is_snapshot
        release_version.increment(release_type)
        return release_version

    def create_bump_version(self):
        bump_version = Version(self.id, self.version_list.copy())
        bump_version.is_snapshot = self.is_snapshot
        bump_version.increment(ReleaseType.BUMP)
        return bump_version

class Release():
    def __init__(self, release_type: ReleaseType, version: Version, current_branch: str):
        self.release_type = release_type
        self.version = version
        self.current_branch = current_branch

    def release_version(self):
        return self.version.create_release_version(self.release_type)

    def bump_version(self):
        return self.release_version().create_bump_version()

    def validate(self, main_branch):
        result = []
        if self.release_type is not None and main_branch != self.current_branch:
            result.append(f"Releases are allowed only on {main_branch}")
        return result

    def is_valid(self, main_branch):
        return self.validate(main_branch).count < 1



import deprecation
from subprocess import run, CalledProcessError
from .domain import Devops
from .infrastructure import ProjectRepository, FileApi


@deprecation.deprecated(deprecated_in="3.2", details="create objects direct instead")
def create_devops_build_config(
    stage, project_root_path, module, build_dir_name="target"
):
    return {
        "stage": stage,
        "project_root_path": project_root_path,
        "module": module,
        "build_dir_name": build_dir_name,
    }

def get_devops_build(project):
    return project.get_property("devops_build")


@deprecation.deprecated(deprecated_in="3.2")
# TODO: Remove from here!
def get_tag_from_latest_commit():
    try:
        value = run(
            "git describe --abbrev=0 --tags --exact-match",
            shell=True,
            capture_output=True,
            check=True,
        )
        return value.stdout.decode("UTF-8").rstrip()
    except CalledProcessError:
        return None


class DevopsBuild:
    def __init__(self, project, config: map = None, devops: Devops = None):
        self.project = project
        self.file_api = FileApi()
        self.repo = ProjectRepository()
        if not devops:
            devops = Devops(
                stage=config["stage"],
                project_root_path=config["project_root_path"],
                module=config["module"],
                name=project.name,
                build_dir_name=config["build_dir_name"],
            )
        self.repo.set_devops(self.project, devops)
        self.repo.set_build(self.project, self)

    def name(self):
        devops = self.repo.get_devops(self.project)
        return devops.name

    def build_path(self):
        devops = self.repo.get_devops(self.project)
        return devops.build_path()

    def initialize_build_dir(self):
        devops = self.repo.get_devops(self.project)
        self.file_api.clean_dir(devops.build_path())



from .devops_terraform_build import DevopsTerraformBuild


def add_exoscale_mixin_config(config, exoscale_api_key, exoscale_secret_key):
    config.update({'ExoscaleMixin':
                   {'exoscale_api_key': exoscale_api_key,
                    'exoscale_secret_key': exoscale_secret_key}})
    return config


class ExoscaleMixin(DevopsTerraformBuild):

    def __init__(self, project, config):
        super().__init__(project, config)
        exoscale_mixin_config = config['ExoscaleMixin']
        self.exoscale_api_key = exoscale_mixin_config['exoscale_api_key']
        self.exoscale_secret_key = exoscale_mixin_config['exoscale_secret_key']

    def project_vars(self):
        ret = super().project_vars()
        if self.exoscale_api_key:
            ret['exoscale_api_key'] = self.exoscale_api_key
        if self.exoscale_secret_key:
            ret['exoscale_secret_key'] = self.exoscale_secret_key
        return ret

    def copy_build_resources_from_package(self):
        super().copy_build_resources_from_package()
        self.copy_build_resource_file_from_package('provider_registry.tf')
        self.copy_build_resource_file_from_package('exoscale_provider.tf')
        self.copy_build_resource_file_from_package('exoscale_mixin_vars.tf')



import json
import re
import subprocess as sub
from abc import ABC, abstractmethod
from pathlib import Path


class FileHandler(ABC):

    @classmethod
    def from_file_path(cls, file_path):
        config_file_type = file_path.suffix
        match config_file_type:
            case '.json':
                file_handler = JsonFileHandler()
            case '.gradle':
                file_handler = GradleFileHandler()
            case '.clj':
                file_handler = ClojureFileHandler()
            case '.py':
                file_handler = PythonFileHandler()
            case _:
                raise Exception(
                    f'The file type "{config_file_type}" is not implemented')

        file_handler.config_file_path = file_path
        file_handler.config_file_type = config_file_type
        return file_handler

    @abstractmethod
    def parse(self) -> tuple[list[int], bool]:
        pass

    @abstractmethod
    def write(self, version_string):
        pass


class JsonFileHandler(FileHandler):

    def parse(self) -> tuple[list[int], bool]:
        with open(self.config_file_path, 'r') as json_file:
            json_version = json.load(json_file)['version']
            is_snapshot = False
            if '-SNAPSHOT' in json_version:
                is_snapshot = True
                json_version = json_version.replace('-SNAPSHOT', '')
            version = [int(x) for x in json_version.split('.')]
            return version, is_snapshot

    def write(self, version_string):
        with open(self.config_file_path, 'r+') as json_file:
            json_data = json.load(json_file)
            json_data['version'] = version_string
            json_file.seek(0)
            json.dump(json_data, json_file, indent=4)
            json_file.truncate()


class GradleFileHandler(FileHandler):

    def parse(self) -> tuple[list[int], bool]:
        with open(self.config_file_path, 'r') as gradle_file:
            contents = gradle_file.read()
            version_line = re.search("\nversion = .*", contents)
            exception = Exception("Version not found in gradle file")
            if version_line is None:
                raise exception

            version_line = version_line.group()
            version_string = re.search(
                '[0-9]*\\.[0-9]*\\.[0-9]*(-SNAPSHOT)?', version_line)
            if version_string is None:
                raise exception

            version_string = version_string.group()
            is_snapshot = False
            if '-SNAPSHOT' in version_string:
                is_snapshot = True
                version_string = version_string.replace('-SNAPSHOT', '')

            version = [int(x) for x in version_string.split('.')]

            return version, is_snapshot

    def write(self, version_string):
        with open(self.config_file_path, 'r+') as gradle_file:
            contents = gradle_file.read()
            version_substitute = re.sub(
                '\nversion = "[0-9]*\\.[0-9]*\\.[0-9]*(-SNAPSHOT)?"', f'\nversion = "{version_string}"', contents)
            gradle_file.seek(0)
            gradle_file.write(version_substitute)
            gradle_file.truncate()


class PythonFileHandler(FileHandler):

    def parse(self) -> tuple[list[int], bool]:
        with open(self.config_file_path, 'r') as python_file:
            contents = python_file.read()
            version_line = re.search("\nversion = .*\n", contents)
            exception = Exception("Version not found in gradle file")
            if version_line is None:
                raise exception

            version_line = version_line.group()
            version_string = re.search(
                '[0-9]*\\.[0-9]*\\.[0-9]*(-SNAPSHOT)?', version_line)
            if version_string is None:
                raise exception

            version_string = version_string.group()
            is_snapshot = False
            if '-SNAPSHOT' in version_string:
                is_snapshot = True
                version_string = version_string.replace('-SNAPSHOT', '')

            version = [int(x) for x in version_string.split('.')]

            return version, is_snapshot

    def write(self, version_string):
        with open(self.config_file_path, 'r+') as python_file:
            contents = python_file.read()
            version_substitute = re.sub(
                '\nversion = "[0-9]*\\.[0-9]*\\.[0-9]*(-SNAPSHOT)?"', f'\nversion = "{version_string}"', contents)
            python_file.seek(0)
            python_file.write(version_substitute)
            python_file.truncate()


class ClojureFileHandler(FileHandler):

    def parse(self) -> tuple[list[int], bool]:
        with open(self.config_file_path, 'r') as clj_file:
            contents = clj_file.read()
            version_line = re.search("^\\(defproject .*\n", contents)
            exception = Exception("Version not found in clj file")
            if version_line is None:
                raise exception

            version_line = version_line.group()
            version_string = re.search(
                '[0-9]*\\.[0-9]*\\.[0-9]*(-SNAPSHOT)?', version_line)
            if version_string is None:
                raise exception

            version_string = version_string.group()
            is_snapshot = False
            if '-SNAPSHOT' in version_string:
                is_snapshot = True
                version_string = version_string.replace('-SNAPSHOT', '')

            version = [int(x) for x in version_string.split('.')]

            return version, is_snapshot

    def write(self, version_string):
        with open(self.config_file_path, 'r+') as clj_file:
            clj_first = clj_file.readline()
            clj_rest = clj_file.read()
            version_substitute = re.sub(
                '[0-9]*\\.[0-9]*\\.[0-9]*(-SNAPSHOT)?', f'"{version_string}"\n', clj_first)
            clj_file.seek(0)
            clj_file.write(version_substitute)
            clj_file.write(clj_rest)
            clj_file.truncate()


class SystemAPI():

    def __init__(self):
        self.stdout = [""]
        self.stderr = [""]

    def run(self, args):
        stream = sub.Popen(args,
                           stdout=sub.PIPE,
                           stderr=sub.PIPE,
                           text=True,
                           encoding="UTF-8")
        self.stdout = stream.stdout.readlines()
        self.stderr = stream.stderr.readlines()

    def run_checked(self, *args):
        self.run(args)

        if len(self.stderr) > 0:
            raise Exception(f"Command failed with: {self.stderr}")


class GitApi():

    def __init__(self):
        self.system_api = SystemAPI()

    def get_latest_n_commits(self, n: int):
        self.system_api.run_checked(
            'git', 'log', '--oneline', '--format="%s %b"', f'-n {n}')
        return self.system_api.stdout

    def get_latest_commit(self):
        output = self.get_latest_n_commits(1)
        return " ".join(output)

    def tag_annotated(self, annotation: str, message: str, count: int):
        self.system_api.run_checked(
            'git', 'tag', '-a', annotation, '-m', message, f'HEAD~{count}')
        return self.system_api.stdout

    def get_latest_tag(self):
        self.system_api.run_checked('git', 'describe', '--tags', '--abbrev=0')
        return self.system_api.stdout

    def get_current_branch(self):
        self.system_api.run_checked('git', 'branch', '--show-current')
        return ''.join(self.system_api.stdout).rstrip()

    def init(self):
        self.system_api.run_checked('git', 'init')

    def add_file(self, file_path: Path):
        self.system_api.run_checked('git', 'add', file_path)
        return self.system_api.stdout

    def commit(self, commit_message: str):
        self.system_api.run_checked(
            'git', 'commit', '-m', commit_message)
        return self.system_api.stdout

    def push(self):
        self.system_api.run_checked('git', 'push')
        return self.system_api.stdout

    def checkout(self, branch: str):
        self.system_api.run_checked('git', 'checkout', branch)
        return self.system_api.stdout



import deprecation
from .domain import Docker
from .application import DockerBuildService
from .devops_build import DevopsBuild, create_devops_build_config


@deprecation.deprecated(deprecated_in="3.2", details="create objects direct instead")
def create_devops_docker_build_config(
    stage,
    project_root_path,
    module,
    dockerhub_user,
    dockerhub_password,
    build_dir_name="target",
    use_package_common_files=True,
    build_commons_path=None,
    docker_build_commons_dir_name="docker",
    docker_publish_tag=None,
):
    ret = create_devops_build_config(stage, project_root_path, module, build_dir_name)
    ret.update(
        {
            "dockerhub_user": dockerhub_user,
            "dockerhub_password": dockerhub_password,
            "use_package_common_files": use_package_common_files,
            "docker_build_commons_dir_name": docker_build_commons_dir_name,
            "build_commons_path": build_commons_path,
            "docker_publish_tag": docker_publish_tag,
        }
    )
    return ret


class DevopsDockerBuild(DevopsBuild):
    def __init__(self, project, config: map = None, docker: Docker = None):
        self.docker_build_service = DockerBuildService()
        if not docker:
            docker = Docker(
                dockerhub_user=config["dockerhub_user"],
                dockerhub_password=config["dockerhub_password"],
                use_package_common_files=config["use_package_common_files"],
                build_commons_path=config["build_commons_path"],
                docker_build_commons_dir_name=config["docker_build_commons_dir_name"],
                docker_publish_tag=config["docker_publish_tag"],
            )
            super().__init__(project, config=config)
        else:
            super().__init__(project, devops=docker.devops)
        self.repo.set_docker(self.project, docker)

    def initialize_build_dir(self):
        super().initialize_build_dir()
        docker = self.repo.get_docker(self.project)
        self.docker_build_service.initialize_build_dir(docker)

    def image(self):
        docker = self.repo.get_docker(self.project)
        self.docker_build_service.image(docker)

    def drun(self):
        docker = self.repo.get_docker(self.project)
        self.docker_build_service.drun(docker)

    def dockerhub_login(self):
        docker = self.repo.get_docker(self.project)
        self.docker_build_service.dockerhub_login(docker)

    def dockerhub_publish(self):
        docker = self.repo.get_docker(self.project)
        self.docker_build_service.dockerhub_publish(docker)

    def test(self):
        docker = self.repo.get_docker(self.project)
        self.test.dockerhub_publish(docker)



from dda_python_terraform import Terraform
from .digitalocean_terraform_build import DigitaloceanTerraformBuild

def add_digitalocean_backend_properties_mixin_config(config,
                                                     account_name,
                                                     endpoint,
                                                     bucket,
                                                     key,
                                                     region='eu-central-1'):
    config.update({'DigitaloceanBackendPropertiesMixin':
                   {'account_name': account_name,
                    'endpoint': endpoint,
                    'bucket': bucket,
                    'key': key,
                    'region': region}})
    return config


class DigitaloceanBackendPropertiesMixin(DigitaloceanTerraformBuild):

    def __init__(self, project, config):
        super().__init__(project, config)
        do_mixin_config = config['DigitaloceanBackendPropertiesMixin']
        self.account_name = do_mixin_config['account_name']
        self.endpoint = do_mixin_config['endpoint']
        self.bucket = do_mixin_config['bucket']
        self.key = do_mixin_config['account_name'] + \
            '/' + do_mixin_config['key']
        self.region = do_mixin_config['region']
        self.backend_config = {
            'access_key': self.do_spaces_access_id,
            'secret_key': self.do_spaces_secret_key,
            'endpoint': self.endpoint,
            'bucket': self.bucket,
            'key': self.key,
            'region': self.region}

    def project_vars(self):
        ret = super().project_vars()
        ret.update({'account_name': self.account_name})
        ret.update({'endpoint': self.endpoint})
        ret.update({'bucket': self.bucket})
        ret.update({'key': self.key})
        ret.update({'region': self.region})
        return ret

    def copy_build_resources_from_package(self):
        super().copy_build_resources_from_package()
        self.copy_build_resource_file_from_package(
            'do_backend_properties_vars.tf')
        self.copy_build_resource_file_from_package(
            'do_backend_with_properties.tf')

    def copy_local_state(self):
        pass

    def rescue_local_state(self):
        pass

    def init_client(self):
        terraform = Terraform(working_dir=self.build_path(),
                       terraform_semantic_version=self.terraform_semantic_version)
        terraform.init(backend_config=self.backend_config)
        self.print_terraform_command(terraform)
        if self.use_workspace:
            try:
                terraform.workspace('select', self.stage)
                self.print_terraform_command(terraform)
            except:
                terraform.workspace('new', self.stage)
                self.print_terraform_command(terraform)
        return terraform



from .domain import Devops, Docker
from .infrastructure import FileApi, ResourceApi, DockerApi, ExecutionApi


class DockerBuildService:
    def __init__(self):
        self.file_api = FileApi()
        self.resource_api = ResourceApi()
        self.docker_api = DockerApi()

    def __copy_build_resource_file_from_package__(self, resource_name, docker: Docker):
        data = self.resource_api.read_resource(f"src/main/resources/docker/{resource_name}")
        self.file_api.write_data_to_file(
            f"{docker.devops.build_path()}/{resource_name}", data
        )

    def __copy_build_resources_from_package__(self, docker: Docker):
        self.__copy_build_resource_file_from_package__(
            "image/resources/install_functions.sh", docker
        )

    def __copy_build_resources_from_dir__(self, docker: Docker):
        self.file_api.cp_force(
            docker.docker_build_commons_path(), docker.devops.build_path()
        )

    def initialize_build_dir(self, docker: Docker):
        build_path = docker.devops.build_path()
        self.file_api.clean_dir(f"{build_path}/image/resources")
        if docker.use_package_common_files:
            self.__copy_build_resources_from_package__(docker)
        else:
            self.__copy_build_resources_from_dir__(docker)
        self.file_api.cp_recursive("image", build_path)
        self.file_api.cp_recursive("test", build_path)

    def image(self, docker: Docker):
        self.docker_api.image(docker.devops.name, docker.devops.build_path())

    def drun(self, docker: Devops):
        self.docker_api.drun(docker.devops.name)

    def dockerhub_login(self, docker: Docker):
        self.docker_api.dockerhub_login(
            docker.dockerhub_user, docker.dockerhub_password
        )

    def dockerhub_publish(self, docker: Docker):
        self.docker_api.dockerhub_publish(
            docker.devops.name, docker.dockerhub_user, docker.docker_publish_tag
        )

    def test(self, docker: Docker):
        self.docker_api.test(docker.devops.name, docker.devops.build_path())



from pathlib import Path
from sys import stdout
from pkg_resources import resource_string
from os import chmod
import yaml
from .domain import Devops, Docker, C4k
from .python_util import execute, execute_live


class ProjectRepository:
    def set_build(self, project, build):
        project.set_property("devops_build", build)

    def get_devops(self, project) -> Devops:
        return project.get_property("build")

    def set_devops(self, project, build: Devops):
        project.set_property("build", build)

    def  get_docker(self, project) -> Docker:
        return project.get_property("docker_build")

    def  set_docker(self, project, build: Docker):
        project.set_property("docker_build", build)

    def get_c4k(self, project) -> C4k:
        return project.get_property("c4k_build")

    def set_c4k(self, project, build: C4k):
        project.set_property("c4k_build", build)


class ResourceApi:
    def read_resource(self, path: str) -> bytes:
        return resource_string(__name__, path)


class FileApi:
    def clean_dir(self, directory: str):
        execute("rm -rf " + directory, shell=True)
        execute("mkdir -p " + directory, shell=True)

    def cp_force(self, src: str, target_dir: str):
        execute("cp -f " + src + "* " + target_dir, shell=True)

    def cp_recursive(self, src: str, target_dir: str):
        execute("cp -r " + src + " " + target_dir, shell=True)

    def write_data_to_file(self, path: Path, data: bytes):
        with open(path, "w", encoding="utf-8") as output_file:
            output_file.write(data.decode(stdout.encoding))

    def write_yaml_to_file(self, path: Path, data: map):
        with open(path, "w", encoding="utf-8") as output_file:
            yaml.dump(data, output_file)
        chmod(path, 0o600)


class DockerApi:
    def image(self, name: str, path: Path):
        execute_live(
            "docker build -t "
            + name
            + " --file "
            + path
            + "/image/Dockerfile "
            + path
            + "/image",
            shell=True,
        )

    def drun(self, name: str):
        execute('docker run -it --entrypoint="" ' + name + " /bin/bash", shell=True)

    def dockerhub_login(self, username: str, password: str):
        execute(
            "docker login --username " + username + " --password " + password,
            shell=True,
        )

    def dockerhub_publish(self, name: str, username: str, tag=None):
        if tag is not None:
            execute(
                "docker tag " + name + " " + username + "/" + name + ":" + tag,
                shell=True,
            )
            execute("docker push " + username + "/" + name + ":" + tag, shell=True)
        execute(
            "docker tag " + name + " " + username + "/" + name + ":latest", shell=True
        )
        execute("docker push " + username + "/" + name + ":latest", shell=True)

    def test(self, name: str, path: Path):
        execute(
            "docker build -t "
            + name
            + "-test "
            + "--file "
            + path
            + "/test/Dockerfile "
            + path
            + "/test",
            shell=True,
        )


class ExecutionApi:
    def execute(command: str, dry_run=False):
        output = ""
        if dry_run:
            print(command)
        else:
            output = execute(command, True)
            print(output)
        return output



from subprocess import check_output, Popen, PIPE
import sys

def execute(cmd, shell=False):
    output = check_output(cmd, encoding='UTF-8', shell=shell)
    return output.rstrip()

def execute_live(cmd):
    process = Popen(cmd, stdout=PIPE)
    for line in iter(process.stdout.readline, b''):
        print(line.decode('utf-8'), end='')
    process.stdout.close()
    process.wait()

def filter_none(list_to_filter):
    return [x for x in list_to_filter if x is not None]
